from torch import nn

# from layers.backbones.lss_fpn import LSSFPN
from layers.backbones.uda_fpn import LSSFPN
from layers.heads.bev_depth_head import BEVDepthHead

__all__ = ['BEVDepth']


class BEVDepth(nn.Module):
    """Source code of `BEVDepth`, `https://arxiv.org/abs/2112.11790`.

    Args:
        backbone_conf (dict): Config of backbone.
        head_conf (dict): Config of head.
        is_train_depth (bool): Whether to return depth.
            Default: False.
    """

    # TODO: Reduce grid_conf and data_aug_conf
    def __init__(self, backbone_conf, head_conf, is_train_depth=False):
        super(BEVDepth, self).__init__()
        self.backbone = LSSFPN(**backbone_conf)
        self.head = BEVDepthHead(**head_conf)
        self.is_train_depth = is_train_depth

    def forward(
        self,
        x,
        mats_dict,
        depth_label=None,
        timestamps=None,
    ):
        """Forward function for BEVDepth

        Args:
            x (Tensor): Input ferature map.
            mats_dict(dict):
                sensor2ego_mats(Tensor): Transformation matrix from
                    camera to ego with shape of (B, num_sweeps,
                    num_cameras, 4, 4).
                intrin_mats(Tensor): Intrinsic matrix with shape
                    of (B, num_sweeps, num_cameras, 4, 4).
                ida_mats(Tensor): Transformation matrix for ida with
                    shape of (B, num_sweeps, num_cameras, 4, 4).
                sensor2sensor_mats(Tensor): Transformation matrix
                    from key frame camera to sweep frame camera with
                    shape of (B, num_sweeps, num_cameras, 4, 4).
                bda_mat(Tensor): Rotation matrix for bda with shape
                    of (B, 4, 4).
            timestamps (long): Timestamp.
                Default: None.

        Returns:
            tuple(list[dict]): Output results for tasks.
        """
        if self.is_train_depth and self.training:
            if depth_label is None:
                x, depth_pred, img_feats, voxel_feats, camera_feats = self.backbone(x,
                                        mats_dict,
                                        depth_label=None,
                                        timestamps=timestamps,
                                        is_return_depth=True)
            else:
                # print("Your are forwarding with out depthNet")
                x, depth_pred, img_feats, voxel_feats, camera_feats = self.backbone(x,
                                            mats_dict,
                                            depth_label,
                                            timestamps,
                                            is_return_depth=True)
            preds = self.head(x)
            return preds, depth_pred, img_feats, x, voxel_feats, camera_feats
        else:
            if depth_label is None:
                x, depth_pred, img_feats, voxel_feats, camera_feats = self.backbone(x,
                                        mats_dict,
                                        depth_label=None,
                                        timestamps=timestamps,
                                        is_return_depth=True)
            else:
                # print("Your are forwarding with out depthNet")
                x, depth_pred, img_feats, voxel_feats, camera_feats = self.backbone(x,
                                            mats_dict,
                                            depth_label,
                                            timestamps,
                                            is_return_depth=True)
            # origin eval code here
            # if depth_label is None:
            #     x = self.backbone(x, mats_dict, timestamps)
            # else:
            #     x = self.backbone(x,
            #                     mats_dict,
            #                     depth_label=depth_label,
            #                     timestamps=timestamps)
            preds = self.head(x)
            return preds, depth_pred, img_feats, x, voxel_feats, camera_feats

    def get_targets(self, gt_boxes, gt_labels):
        """Generate training targets for a single sample.

        Args:
            gt_bboxes_3d (:obj:`LiDARInstance3DBoxes`): Ground truth gt boxes.
            gt_labels_3d (torch.Tensor): Labels of boxes.

        Returns:
            tuple[list[torch.Tensor]]: Tuple of target including \
                the following results in order.

                - list[torch.Tensor]: Heatmap scores.
                - list[torch.Tensor]: Ground truth boxes.
                - list[torch.Tensor]: Indexes indicating the position \
                    of the valid boxes.
                - list[torch.Tensor]: Masks indicating which boxes \
                    are valid.
        """
        return self.head.get_targets(gt_boxes, gt_labels)

    def loss(self, targets, preds_dicts):
        """Loss function for BEVDepth.

        Args:
            gt_bboxes_3d (list[:obj:`LiDARInstance3DBoxes`]): Ground
                truth gt boxes.
            gt_labels_3d (list[torch.Tensor]): Labels of boxes.
            preds_dicts (dict): Output of forward function.

        Returns:
            dict[str:torch.Tensor]: Loss of heatmap and bbox of each task.
        """
        return self.head.loss(targets, preds_dicts)

    def get_bboxes(self, preds_dicts, img_metas=None, img=None, rescale=False):
        """Generate bboxes from bbox head predictions.

        Args:
            preds_dicts (tuple[list[dict]]): Prediction results.
            img_metas (list[dict]): Point cloud and image's meta info.

        Returns:
            list[dict]: Decoded bbox, scores and labels after nms.
        """
        return self.head.get_bboxes(preds_dicts, img_metas, img, rescale)
